/*
 * Copyright (c) 1992, 1993, 1994, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * This software was developed by the Computer Systems Engineering group
 * at Lawrence Berkeley Laboratory under DARPA contract BG 91-66 and
 * contributed to Berkeley.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Lawrence Berkeley Laboratory and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) $Header: bootbootp.c,v 1.24 95/04/12 02:16:00 leres Exp $ (LBL)
 */

#include <sys/param.h>
#include <sys/mount.h>
#include <sys/socket.h>

#include <net/if.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>

#include <nfs/nfsv2.h>
#include <nfs/nfsdiskless.h>

#include <arpa/inet.h>

#include <a.out.h>
#include <errno.h>
#ifdef TEST
#include <fcntl.h>
#include <unistd.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef LBL
#include <lbl/os-proto.h>
#endif

#include "bootbootp.h"
#include "bootp.h"

#define BASEADDR 0x4000

/* Globals */
u_char	bcea[6] = BA;			/* broadcast ethernet address */
char	rootpath[FNAME_SIZE];		/* root mount path */
char	swappath[FNAME_SIZE];		/* swap mount path */
char	ifname[IFNAME_SIZE];		/* name of interface (e.g. "le0") */
n_long	rootip;				/* root ip address */
n_long	swapip;				/* swap ip address */
n_long	gateip;				/* swap ip address */
n_long	smask;				/* subnet mask */
n_long	nmask;				/* net mask */
n_long	mask;				/* subnet or net mask */
time_t	bot;				/* beginning of time in seconds */

int	nflag = 0;

/* Locals */
static	u_char rootfh[NFS_FHSIZE];
static	u_char swapfh[NFS_FHSIZE];
static	u_long swapblks;
static	time_t roottime;
#ifdef TEST
static	int td;
#endif

/* Local forwards */
#if !defined(TEST) && defined(CKSUM)
static	void cksum __P((u_int *, u_int *));
#endif
static	void diskless_setup __P((struct iodesc *, struct nfs_diskless_v1 *));
static	u_long findsym __P((struct iodesc *, const char *, struct exec *));
static	char *intoa __P((n_long));
int	main __P((int, char **));
static	void readseg __P((struct iodesc *, u_long, u_long, u_long));


#ifndef TEST
extern	struct promvec *promvec;
#endif

int
main(argc, argv)
	int argc;
	char **argv;
{
	register int i, cc;
	register u_long sym, txtaddr, tenths;
	u_long bytes, ftype;
	register struct iodesc *d;
	register char *name;
	struct iodesc iodesc;
	struct exec exec;
	u_char image_fh[NFS_FHSIZE];
	char localpath[sizeof(swappath)];
	char *image;

	bot = getsecs();
	printf("%s loaded and running...\n", version);

	d = &iodesc;
	bzero(d, sizeof(*d));

	/* Setup house */
	ethernet_init(d);

	--argc;
	++argv;
	image = "vmunix";		/* default image */
	if (argc > 0) {
		image = *argv;
		--argc;
		++argv;
	}

	ethernet_macaddr(d, sizeof(d->myea), d->myea);

	/* Base transaction id on ethernet address */
	bcopy(&d->myea[2], &d->xid, 4);

	/* Get bootp info */
	bootp(d);

	/*
	 * Now that bootp is done, base myport on ethernet address Note
	 * that many nfs servers require a privileged port so we insure:
	 *
	 *	IPPORT_RESERVED / 2 <= port < IPPORT_RESERVED
	 */
	bcopy(&d->myea[4], &d->myport, 2);
	d->myport = IPPORT_RESERVED / 2 + (d->myport % IPPORT_RESERVED / 2);

	printf("myip: %s", intoa(d->myip));
	if (gateip)
		printf(", gateip: %s", intoa(gateip));
	if (mask)
		printf(", mask: %s", intoa(mask));
	printf("\n");
	printf("root: %s:%s\n", intoa(rootip), rootpath);
	printf("swap: %s:%s\n", intoa(swapip), swappath);

	/* Get root file handle and timestamp */
	d->destip = rootip;
	getnfsfh(d, rootpath, rootfh);
	d->fh = rootfh;
	getnfsinfo(d, &roottime, NULL, &ftype);
	if (ftype != NFDIR)
		printf("warning: bad root ftype %lu\n", ftype);
#ifdef notdef
	printf("root fs time: %lu\n", roottime);
#endif

	/* Get image file handle and size */
	if (lookupfh(d, image, image_fh, NULL, &bytes, &ftype) < 0)
		panic("bootbootp: lookup %s: %m", image);
	if (ftype != NFREG)
		printf("warning: bad image ftype %lu\n", ftype);
#ifdef TEST
	name = "a.out";
	td = open(name, O_WRONLY | O_TRUNC, 0666);
	if (td < 0)
		printf("bootbootp: open %s: %m\n", name);
#endif
	d->fh = image_fh;
	cc = readdata(d, 0, &exec, sizeof(exec));
	if (cc != sizeof(exec)) {
		if (cc < 0)
			panic("bootbootp: %s exec read: %m", image);
		panic("bootbootp: bad byte count (%d != %d)", cc, sizeof(exec));
	}
	if (exec.a_machtype != MID_SUN_SPARC)
		panic("bootbootp: bad machine type %d", exec.a_machtype);
	if (N_BADMAG(exec))
		panic("bootbootp: bad magic 0%o", exec.a_magic);

#ifdef TEST
	if (td >= 0 && write(td, &exec, sizeof(exec)) != sizeof(exec))
		panic("bootbootp: exec write: %m");
#endif

	/* Get swap file handle and size */
	d->destip = swapip;
	getnfsfh(d, swappath, swapfh);
	d->fh = swapfh;
	getnfsinfo(d, NULL, &bytes, &ftype);
	if (ftype != NFREG)
		printf("warning: bad swap ftype %lu\n", ftype);
	/*
	 * If the swap file is less than 512 bytes in length, then it
	 * contains the name of our local swap device.
	 */
	if (bytes < 512) {
		i = sizeof(localpath) - 1;
		if (i > bytes)
			i = bytes;
		cc = readdata(d, 0, localpath, i);
		if (cc < 0)
			panic("read local swap file %s: %m", swappath);
		if (cc > 0 && localpath[cc - 1] == '\n')
			--cc;
		localpath[cc] = '\0';
		(void)strcpy(swappath, localpath);

		/* Indicate local swap device */
		swapip = htonl(0x7f000001);
		swapblks = 0;

		/* Clobber swap file handle */
		bzero(swapfh, sizeof(swapfh));

		printf("swap: %s (local)\n", swappath);
	} else {
		if (bytes % 512)
			printf("warning: swap is odd sized (%lu bytes)\n",
			    bytes);
		swapblks = bytes / 512;

		printf("swap: %luK\n", bytes / 1024);
	}

	/* Read image into memory */
	tenths = gettenths();
	printf("%s: %lu", image, exec.a_text);
	d->fh = image_fh;
#ifdef notdef
	txtaddr = exec.a_entry - KERNBASE;
#else
	txtaddr = BASEADDR;
#endif
	readseg(d, N_TXTOFF(exec), exec.a_text, txtaddr);
	printf("+%lu", exec.a_data);
	readseg(d, N_DATOFF(exec), exec.a_data, txtaddr + exec.a_text);

	/* The kernel bzero's bss so we don't have to */
	printf("+%lu\n", exec.a_bss);

	tenths = gettenths() - tenths;
	printf("[took %lu.%lu seconds", tenths / 10, tenths % 10);
	if (tenths) {
		/* calculate kbytes/sec */
		i = (exec.a_text + exec.a_data) / tenths;
		i /= 1024;
		printf(", %d.%d kbytes/sec", i / 10, i % 10);
	}
	printf("]\n");

	/* Read the symbol table */
	name = "_nfs_diskless";
#ifdef notdef
	tenths = gettenths();
#endif
	sym = findsym(d, name, &exec);
#ifdef notdef
	tenths = gettenths() - tenths;
#endif
	if (sym) {
#ifdef notdef
		printf("[\"%s\" at 0x%x, took %d.%d seconds]\n",
		    name, sym, tenths / 10, tenths % 10);
#endif
	} else
		printf("bootbootp: warning: can't find symbol \"%s\"\n", name);
#ifdef TEST
	if (sym)
		sym = (u_long)malloc(sizeof(struct nfs_diskless_v1));
#endif

	/* Now we're ready to fill in the nfs_diskless struct */
	if (sym) {
		if (nflag)
			printf("bootbootp: skipping nfs_diskless setup\n");
		else {
			/* Assume kernel not larger than 128 Mbytes */
			sym &= ~0xf8000000;
			diskless_setup(d, (struct nfs_diskless_v1 *)sym);
		}
	}

	/* Shutdown the ethernet */
	ethernet_shutdown(d);

#ifndef TEST
#ifdef CKSUM
	{
	u_long s, e;

	s = BASEADDR;
	e = BASEADDR + exec.a_text;
	cksum((u_int *)s, (u_int *)e);
	s = e;
	e += exec.a_data;
	cksum((u_int *)s, (u_int *)e);
	}
#endif
	printf("starting %s...\n", image);
	call(BASEADDR);
	panic("not reached");
#else
	exit(0);
#endif
}

static void
readseg(d, off, size, addr)
	register struct iodesc * d;
	register u_long off, size, addr;
{
	register int i, cc;
#ifdef TEST
	char buf[32 * 1024];

	addr = (u_long)buf;
#endif

	i = 0;
	while (size > 0) {
		cc = readdata(d, off, (void *)addr, size);
		if (cc <= 0) {
			/* XXX maybe should retry on certain errors */
			if (cc < 0)
				panic("\nreadseg: read: %m");
			panic("\nreadseg: hit EOF unexpectantly");
		}
		off += cc;
#ifdef TEST
		if (td >= 0 && write(td, (void *)addr, cc) != cc)
			panic("\nreadseg: write: %m");
#else
		addr += cc;
#endif
		size -= cc;
		i = ~i;
		printf("%c\010", i ? '-' : '=');
	}
}

static char pretty[4] = { '/', '|', '\\', '-' };

static u_long
findsym(d, wanted, ep)
	register struct iodesc *d;
	register const char *wanted;
	register struct exec *ep;
{
	register struct nlist *nl;
	register int nsym, n;
	char *cp;
	int strbase, strend, len;
	int stroff, symoff, symsize, strsize;
	struct nlist symbuf[8192 / sizeof(struct nlist)];
	char strbuf[8*1024];
	register u_long nread;

	len = strlen(wanted) + 1;	/* including NUL */
	symoff = N_SYMOFF(*ep);
	symsize = ep->a_syms;
	stroff = symoff + symsize;

	if (readdata(d, stroff, &strsize, sizeof(strsize)) != sizeof(strsize)) {
		printf("Can't read strsize\n");
		return (0);
	}
	strsize -= sizeof(strsize);

	nread = 0;
	strbase = strend = 0;
	while (symsize >= 0) {
		n = symsize < sizeof(symbuf) ? symsize : sizeof(symbuf);
		n = readdata(d, symoff, symbuf, n);
		++nread;
		printf("%c\010", pretty[nread % 4]);
		nsym = n / sizeof(*nl);
		if (nsym < 1)
			break;
		symsize -= nsym * sizeof(*nl);
		symoff += nsym * sizeof(*nl);
		for (nl = (struct nlist *)symbuf; --nsym >= 0; nl++) {
			if (nl->n_un.n_strx == 0 || nl->n_type & N_STAB)
				continue;
			switch (nl->n_type & ~N_EXT) {

			case N_ABS:
			case N_TEXT:
			case N_DATA:
			case N_BSS:
				break;

			default:
				continue;
			}
			/*
			 * Make sure the string buffer contains the
			 * desired symbol in its entirety.
			 */
			if (nl->n_un.n_strx + len >= strend) {
				strbase = nl->n_un.n_strx & ~7;
				n = strsize - strbase;
				if (n < 0) {
printf("[warning, symbol with bad n_strx]\n");
					continue;
				}
				if (n > sizeof(strbuf))
					n = sizeof(strbuf);
				n = readdata(d, stroff + strbase, strbuf, n);
				++nread;
				printf("%c\010", pretty[nread % 4]);
				if (n < 1) {
					printf("string table read failed\n");
					goto bad;
				}
				strend = strbase + n;
				/* Bail if this string isn't long enough */
				if (nl->n_un.n_strx + len >= strend)
					continue;
			}
			cp = strbuf + (nl->n_un.n_strx - strbase);
			if (strcmp(cp, wanted) == 0)
				return (nl->n_value);
		}
	}
#ifdef notdef
printf("%s: not found\n", wanted);
#endif
bad:
	return (0);
}

static void
diskless_setup(d, nd)
	register struct iodesc *d;
	register struct nfs_diskless_v1 *nd;
{

	nd->version = 1;

	/* My ip address and hostname */
	(void)strcpy(nd->ifname, ifname);
	nd->addr = d->myip;
	strcpy(nd->my_hostnam, intoa(d->myip));

	/* Broadcast address */
	nd->broadaddr = INADDR_BROADCAST;

	/* Subnet (or net) mask */
	nd->mask = mask;

	/* Gateway address */
	if (gateip)
		nd->mygateway = gateip;

	/* Root fileserver */
	nd->root_saddr = rootip;

	(void)strcpy(nd->root_hostnam, intoa(rootip));
	(void)strcat(nd->root_hostnam, ":");
	(void)strcat(nd->root_hostnam, rootpath);
	nd->root_time = roottime;
	bcopy(rootfh, nd->root_fh, sizeof(rootfh));

	/* Swap fileserver */
	nd->swap_saddr = swapip;

	(void)strcpy(nd->swap_hostnam, intoa(swapip));
	(void)strcat(nd->swap_hostnam, ":");
	(void)strcat(nd->swap_hostnam, swappath);
	nd->swap_nblks = swapblks;
	bcopy(swapfh, nd->swap_fh, sizeof(swapfh));

	/* XXX */
#ifdef notdef
	nd->swap_ucred.cr_ref = 1;
	nd->swap_ucred.cr_ngroups = 1;	/* group 0 */
#endif
}

/*
 * Convert Ethernet address to printable (loggible) representation.
 */
static char digits[] = "0123456789abcdef";
char *
ether_sprintf(ap)
        register u_char *ap;
{
	register i;
	static char etherbuf[18];
	register char *cp = etherbuf;

	for (i = 0; i < 6; i++) {
		*cp++ = digits[*ap >> 4];
		*cp++ = digits[*ap++ & 0xf];
		*cp++ = ':';
	}
	*--cp = 0;
	return (etherbuf);
}

/* Similar to inet_ntoa() */
static char *
intoa(addr)
	n_long addr;
{
	register char *cp;
	register u_int byte;
	register int n;
	static char buf[sizeof(".xxx.xxx.xxx.xxx")];

	cp = &buf[sizeof buf];
	*--cp = '\0';

	n = 4;
	do {
		byte = addr & 0xff;
		*--cp = byte % 10 + '0';
		byte /= 10;
		if (byte > 0) {
			*--cp = byte % 10 + '0';
			byte /= 10;
			if (byte > 0)
				*--cp = byte + '0';
		}
		*--cp = '.';
		addr >>= 8;
	} while (--n > 0);

	return (cp + 1);
}

#if !defined(TEST) && defined(CKSUM)
static void
cksum(s, e)
	register u_int *s, *e;
{
	register u_int *p, sum;

	sum = 0;
	for (p = s; p < e; p++)
		sum = ((sum << 1) | (sum >> 31)) ^ *p;
	printf("sum over [%x..%x) = %x\n", (u_long)s, (u_long)e, sum);
}
#endif
