/*
 * Copyright (c) 1992, 1993, 1994, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * This software was developed by the Computer Systems Engineering group
 * at Lawrence Berkeley Laboratory under DARPA contract BG 91-66 and
 * contributed to Berkeley.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Lawrence Berkeley Laboratory and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) $Header: prom_unix.c,v 1.12 95/04/12 02:15:43 leres Exp $ (LBL)
 */

#include <sys/types.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#ifndef	SOCK_STREAM
#include <sys/socket.h>
#endif

#include <netinet/in_systm.h>

#include <net/bpf.h>
#include <net/if.h>

#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define printf Printf
#include <stdio.h>
#undef printf

#ifdef LBL
#include <lbl/os-proto.h>
#endif

#include "bootbootp.h"

/* Locals */
static	char *buffer;
static	int buffersize;

/* Local forwards */
static	void bpf_close __P((int));
static	void bpf_init __P((int, char *));
static	int bpf_open __P((void));
static	int bpf_read __P((int, void *, int, time_t));

static int
bpf_open()
{
	register int n, fd;
	char device[32];

	n = 0;
	do {
		(void)snprintf(device, sizeof(device), "/dev/bpf%d", n);
		if ((fd = open(device, O_RDWR)) >= 0)
			return (fd);
		++n;
	} while (errno == EBUSY);
	return (-1);
}

static void
bpf_init(fd, interface)
	register int fd;
	char *interface;
{
	struct ifreq ifr;
	int linktype;
	int on = 1;

	/* Determine buffer size and allocate */
	if (ioctl(fd, BIOCGBLEN, (caddr_t)&buffersize) < 0)
		panic("bpf_init: BIOCGBLEN: %s", strerror(errno));
	if ((buffer = malloc((unsigned)buffersize)) == 0)
		panic("bpf buffer malloc() failed");

	/* Bind to interface */
	bzero((char *)&ifr, sizeof(ifr));
	(void)strncpy(ifr.ifr_name, interface, sizeof(ifr.ifr_name));
	if (ioctl(fd, BIOCSETIF, (caddr_t)&ifr) < 0)
		panic("bpf_init: BIOCSETIF %s: %s", interface, strerror(errno));

	/* Must be 10 Mbit ethernet */
	if (ioctl(fd, BIOCGDLT, (caddr_t)&linktype) < 0)
		panic("bpf_init: BIOCGDLT %s: %s", interface, strerror(errno));
	if (linktype != DLT_EN10MB)
		panic("bpf_init: %s not 10 Mbit ethernet (#%d)",
		    interface, linktype);

#ifdef notdef
	/* Load filter program */
	/* XXX Could setup to only receive packets to broadcast or us */
	if (ioctl(fd, BIOCSETF, (caddr_t)&aw_prog) < 0)
		panic("bpf_init: BIOCSETF %s", strerror(errno));
#endif

	/* Set immediate mode */
	if (ioctl(fd, BIOCIMMEDIATE, (caddr_t)&on) < 0)
		panic("bpf_init: BIOCIMMEDIATE %s", strerror(errno));
}

/* Returns character count or -1 on timeout. All other errors are fatal */
static int
bpf_read(fd, pkt, len, timeout)
	int fd;
	void *pkt;
	int len;
	time_t timeout;
{
	fd_set fds;
	register int cc;
	struct timeval zero;
	register int caplen, hdrlen, datalen;
	static char *bp = 0, *ep = 0;
#define bhp ((struct bpf_hdr *)bp)

	for (;;) {
		if (bp >= ep) {
			/* Need to read more data */
			FD_ZERO(&fds);
			FD_SET(fd, &fds);
			zero.tv_sec = timeout;
			zero.tv_usec = 0;
			if (select(fd + 1, &fds, NULL, NULL, &zero) < 0)
				panic("bpf_read: select: %s", strerror(errno));

			/* Return if no data yet */
			if (!FD_ISSET(fd, &fds))
				return (-1);

			/* Read data */
			bp = ep = buffer;
			for (;;) {
				cc = read(fd, (char *)bp, buffersize);
				if (cc < 0) {
					if (errno == EINTR)
						continue;
					panic("bpf_read: read: %s",
					    strerror(errno));
				}
				if (cc > buffersize)
					panic("bpf_read: overwrote (%d > %d)",
					    cc, buffersize);
				break;
			}

			/* Update end pointer */
			ep = bp + cc;
		}

		/* Look for next packet */
		while (bp < ep) {
			caplen = bhp->bh_caplen;
			hdrlen = bhp->bh_hdrlen;
			datalen = bhp->bh_datalen;
			if (datalen != caplen) {
				printf("bpf_read: short read! (%d != %d)\n",
				    datalen, caplen);
				bp += BPF_WORDALIGN(hdrlen + caplen);
				continue;
			}
			if (len > caplen)
				len = caplen;
			bcopy(bp + hdrlen, pkt, len);
			bp += BPF_WORDALIGN(hdrlen + caplen);
			return (len);
		}
	}
}

static void
bpf_close(fd)
	register int fd;
{

	(void)close(fd);
}

void
ethernet_init(d)
	register struct iodesc *d;
{

	if ((d->fd = bpf_open()) < 0)
		panic("ethernet_init: bpf_open: %s", strerror(errno));

	if (ifname[0] == '\0')
		(void)strcpy(ifname, "le0");
	bpf_init(d->fd, ifname);
}

void
ethernet_shutdown(d)
	register struct iodesc *d;
{

	bpf_close(d->fd);
	d->fd = 0;
}

/* Returns the actual number of characters written */
int
ethernet_put(d, pkt, len)
	register struct iodesc *d;
	void *pkt;
	int len;
{
	register int cc;

	cc = write(d->fd, pkt, len);
	if (cc < 0)
		panic("ethernet_put: write: %s", strerror(errno));
	return (cc);
}

/* Returns character count or -1 if none. No other error is possible */
int
ethernet_poll(d, pkt, len)
	register struct iodesc *d;
	void *pkt;
	int len;
{
	return (bpf_read(d->fd, pkt, len, 0));
}

/* Returns character count or -1 on timeout. No other error is possible */
int
ethernet_get(d, pkt, len, timeout)
	register struct iodesc *d;
	void *pkt;
	int len;
	time_t timeout;
{
	return (bpf_read(d->fd, pkt, len, timeout));
}

u_long
gettenths()
{
	struct timeval tv;

	(void)gettimeofday(&tv, NULL);
	return (tv.tv_sec * 10 + tv.tv_usec / 100000);
}

time_t
getsecs()
{

	return (time(0));
}

/* Print a formatted message and die */
__dead void
panic(char const *fmt, ...)
{
	register char *cp;
	va_list ap;

	va_start(ap, fmt);
	(void)vfprintf(stderr, fmt, ap);

	/* Output trailing newline if missing in format */
	cp = ((char *)fmt) + strlen(fmt) - 1;
	if (cp >= fmt && *cp != '\n')
		putch('\n');

	va_end(ap);
	exit(1);
}

int
putch(c)
	char c;
{
	return (fputc(c, stderr));
}

int
putstr(s, len)
	const char *s;
	int len;
{
	int n;

	for (n = len; --n >= 0;)
		putch(*s++);
	return (len);
}
